"""
count_assets.py - contains the count_assets function for the Global AssetView API (GAV) module.

for a full list of Qualys QQL filters, see: https://docs.qualys.com/en/gav/2.18.0.0/search/how_to_search.htm
"""

from ..base.call_api import call_api
from ..auth.token import TokenAuth
from ..exceptions.Exceptions import *


def count_assets(auth: TokenAuth, **kwargs) -> dict:
    """
    Count assets in the Global AssetView API based on a QQL filter.

    Params:
        auth (TokenAuth): The authentication object.

    :Kwargs:
        filter (str): The Qualys QQL filter to use.
        lastSeenAssetId (int): The last seen asset ID.
        lastModifiedDate (str): The last modified date.

    Returns:
        dict: The response from the API.
    """
    # despite the fact that this is a POST request, we still need to send stuff as a parameter
    # because Qualys is Qualys.

    # make the request:
    response = call_api(auth=auth, module="gav", endpoint="count_assets", params=kwargs)

    # parse the response:
    return response.json()



"""
Contains the resource name mappings and the class mappings for resources
"""

from .AWSResources import *

VALID_RESOURCETYPES = {
    "aws": [
        "RDS",
        "NETWORK_ACL",
        "BUCKET",
        "IAM_USER",
        "VPC",
        "VPC_SECURITY_GROUP",
        "LAMBDA",
        "SUBNET",
        "INTERNET_GATEWAY",
        "LOAD_BALANCER",
        "EC2_INSTANCE",
        "ROUTE_TABLE",
        "EBS",
        "AUTO_SCALING_GROUP",
        "EKS_CLUSTER",
        "EKS_NODEGROUP",
        "EKS_FARGATE_PROFILE",
        "VPC_ENDPOINT",
        "VPC_ENDPOINT_SERVICE",
        "IAM_GROUP",
        "IAM_POLICY",
        "IAM_ROLE",
        "SAGEMAKER_NOTEBOOK",
        "CLOUDFRONT_DISTRIBUTION",
    ],
    "azure": [
        "SQL_SERVER",
        "FUNCTION_APP",
        "SQL_SERVER_DATABASE",
        "RESOURCE_GROUP",
        "VIRTUAL_NETWORK",
        "VIRTUAL_MACHINE",
        "NETWORK_SECURITY_GROUP",
        "WEB_APP",
        "NETWORK_INTERFACES",
        "POSTGRE_SINGLE_SERVER",
        "LOAD_BALANCER",
        "FIREWALL",
        "MYSQL",
        "STORAGE_ACCOUNT",
        "APPLICATION_GATEWAYS",
        "SECRETS",
        "MARIADB",
        "COSMODB",
        "NAT_GATEWAYS",
    ],
}

# Common names for resources to substitute for the user's convenience
COMMON_NAMES = {
    "aws": {
        "NETWORK_ACL": ["ACL"],
        "BUCKET": ["S3_BUCKET", "S3"],
        "IAM_USER": ["IAM", "IAMUSER"],
        "VPC_SECURITY_GROUP": ["SECURITY_GROUP", "SG"],
        "LAMBDA": ["LAMBDA_FUNCTION"],
        "INTERNET_GATEWAY": ["IG", "IGW", "GATEWAY"],
        "LOAD_BALANCER": ["ELB", "LB"],
        "EC2_INSTANCE": ["EC2", "INSTANCE"],
        "ROUTE_TABLE": ["ROUTE", "ROUTES"],
        "EBS": ["VOLUME", "VOLUMES", "EBS_VOLUME", "EBS_VOLUMES"],
        "AUTO_SCALING_GROUP": ["ASG", "AUTO_SCALING"],
        "EKS_CLUSTER": ["EKS"],
        "EKS_NODEGROUP": ["NODE_GROUP", "EKSNG", "EKS_NODE_GROUP"],
        "EKS_FARGATE_PROFILE": ["FARGATE_PROFILE", "EKS_FARGATE", "FARGATE"],
        "VPC_ENDPOINT_SERVICE": ["ENDPOINT_SERVICE"],
        "IAM_GROUP": ["IAMGROUP"],
        "IAM_POLICY": ["POLICY", "IAMPOLICY"],
        "IAM_ROLE": ["ROLE", "IAMROLE"],
        "SAGEMAKER_NOTEBOOK": ["NOTEBOOK", "SAGEMAKER"],
        "CLOUDFRONT_DISTRIBUTION": ["CLOUDFRONT"],
    },
    "azure": {
        "SQL_SERVER": ["MSSQL"],
        "FUNCTION_APP": ["FUNCTION"],
        "SQL_SERVER_DATABASE": ["MSSQL_DATABASE", "MSSQL_DB", "MSSQLDB"],
        "RESOURCE_GROUP": ["RG"],
        "VIRTUAL_NETWORK": ["VNET"],
        "VIRTUAL_MACHINE": ["VM"],
        "NETWORK_SECURITY_GROUP": ["NSG", "SECURITY_GROUP"],
        "WEB_APP": ["WEBAPP"],
        "NETWORK_INTERFACES": ["NIC", "INTERFACE"],
        "POSTGRE_SINGLE_SERVER": ["POSTGRE", "POSTGRES", "POSTGRESQL"],
        "LOAD_BALANCER": ["LB"],
        "FIREWALL": ["FW"],
        "MYSQL": ["MYSQL_DB", "MYSQLDB"],
        "STORAGE_ACCOUNT": ["STORAGE"],
        "APPLICATION_GATEWAYS": ["AG"],
        "SECRETS": ["SECRET"],
        "MARIADB": ["MARIA_DB", "MARIA"],
        "COSMODB": ["COSMOS", "COSMOS_DB"],
        "NAT_GATEWAYS": ["NAT"],
    },
}

# For dynamically creating the resource object based on the resource type
resource_map = {
    "RDS": AWSRDS,
    "NETWORK_ACL": AWSNetworkACL,
    "BUCKET": AWSBucket,
    "IAM_USER": AWSIAMUser,
    "VPC": AWSVPC,
    "VPC_SECURITY_GROUP": AWSSecurityGroup,
    "LAMBDA": AWSLambdaFunction,
    "SUBNET": AWSSubnet,
    "INTERNET_GATEWAY": AWSInternetGateway,
    "LOAD_BALANCER": AWSLoadBalancer,
    "EC2_INSTANCE": AWSEC2Instance,
    "ROUTE_TABLE": AWSRouteTable,
    "EBS": AWSEBSVolume,
    "AUTO_SCALING_GROUP": AWSAutoScalingGroup,
    "EKS_CLUSTER": AWSEKSCluster,
    "EKS_NODEGROUP": AWSEKSNodeGroup,
    "EKS_FARGATE_PROFILE": AWSEKSFargateProfile,
    "VPC_ENDPOINT": AWSVPCEndpoint,
    "VPC_ENDPOINT_SERVICE": AWSVPCEndpointService,
    "IAM_GROUP": AWSIAMGroup,
    "IAM_POLICY": AWSIAMPolicy,
    "IAM_ROLE": AWSIAMRole,
    "SAGEMAKER_NOTEBOOK": AWSSagemakerNotebook,
    "CLOUDFRONT_DISTRIBUTION": AWSCloudfrontDistribution,
}



"""
vmdr.py - Contains the functions to upload supported VMDR API pulls to SQL DBs.
"""

from datetime import datetime

from pandas import DataFrame
from sqlalchemy import Connection, types
from sqlalchemy.dialects.mysql import TEXT

from .base import upload_data, prepare_dataclass
from ..base.base_list import BaseList


def upload_vmdr_ags(
    ags: BaseList,
    cnxn: Connection,
    table_name: str = "vmdr_assetgroups",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from vmdr.get_ag_list() to SQL.

    Args:
        ags (BaseList): The Asset Group List to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_assetgroups'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "ID": types.Integer(),
        "TITLE": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "OWNER_ID": types.Integer(),
        "OWNER_USER_ID": types.Integer(),
        "UNIT_ID": types.Integer(),
        "LAST_UPDATE": types.DateTime(),
        "NETWORK_ID": types.Integer(),
        "IP_SET": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "BUSINESS_IMPACT": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "DEFAULT_APPLIANCE_ID": types.Integer(),
        "APPLIANCE_IDS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "DNS_LIST": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "NETBIOS_LIST": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "HOST_IDS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "ASSIGNED_USER_IDS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "ASSIGNED_UNIT_IDS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "OWNER_USER_NAME": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CVSS_ENVIRO_CDP": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CVSS_ENVIRO_TD": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CVSS_ENVIRO_CR": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CVSS_ENVIRO_IR": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CVSS_ENVIRO_AR": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "EC2_IDS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "COMMENTS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "DOMAIN_LIST": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CLOUD_GROUP_NAME": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CLOUD_INSTANCE_STATE": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CLOUD_INSTANCE_TYPE": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CLOUD_IS_SPOT_INSTANCE": types.Boolean(),
        "CLOUD_ARCHITECTURE": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CLOUD_REGION": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CLOUD_IMAGE_ID": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CLOUD_AMI_ID": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CLOUD_PUBLIC_HOSTNAME": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CLOUD_PUBLIC_IPV4": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CLOUD_ACCOUNT_ID": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
    }

    df = DataFrame([prepare_dataclass(ag) for ag in ags])

    # Upload the data:
    return upload_data(
        df, table_name, cnxn, dtype=COLS, override_import_dt=override_import_dt
    )


def upload_vmdr_kb(
    kbs: BaseList,
    cnxn: Connection,
    table_name: str = "vmdr_knowledgebase",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from vmdr.query_kb() to SQL.

    Args:
        kbs (BaseList): The KB List to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_knowledgebase'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "QID": types.Integer(),
        "VULN_TYPE": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "SEVERITY_LEVEL": types.Integer(),
        "TITLE": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "CATEGORY": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "LAST_SERVICE_MODIFICATION_DATETIME": types.DateTime(),
        "PUBLISHED_DATETIME": types.DateTime(),
        "CODE_MODIFIED_DATETIME": types.DateTime(),
        "BUGTRAQ_LIST": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "PATCHABLE": types.Boolean(),
        "SOFTWARE_LIST": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "VENDOR_REFERENCE_LIST": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CVE_LIST": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "DIAGNOSIS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "SOLUTION": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CONSEQUENCE": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CORRELATION": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CVSS": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "CVSS_V3": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "PCI_REASONS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "PCI_FLAG": types.Boolean(),
        "THREAT_INTELLIGENCE": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "SUPPORTED_MODULES": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "DISCOVERY": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "IS_DISABLED": types.Boolean(),
        "CHANGE_LOG": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "COMPLIANCE_LIST": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "TECHNOLOGY": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "LAST_CUSTOMIZATION": types.DateTime(),
        "SOLUTION_COMMENT": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
    }

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(kb) for kb in kbs])

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_vmdr_hosts(
    hosts: BaseList,
    cnxn: Connection,
    table_name: str = "vmdr_hosts_list",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from vmdr.get_host_list() to SQL.

    Args:
        hosts (BaseList): The Host List to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_hosts_list'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.

    """

    COLS = {
        "ID": types.Integer(),
        "ASSET_ID": types.Integer(),
        "IP": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "IPV6": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "TRACKING_METHOD": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "DNS": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "NETBIOS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "OS": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "QG_HOSTID": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "LAST_BOOT": types.DateTime(),
        "LAST_SCAN_DATETIME": types.DateTime(),
        "SERIAL_NUMBER": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "HARDWARE_UUID": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "FIRST_FOUND_DATE": types.DateTime(),
        "LAST_ACTIVITY": types.DateTime(),
        "LAST_ACTIVITY_DATE": types.DateTime(),
        "AGENT_STATUS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CLOUD_AGENT_RUNNING_ON": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "TAGS": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "LAST_VULN_SCAN_DATETIME": types.DateTime(),
        "LAST_VULN_SCAN_DATE": types.DateTime(),
        "LAST_VM_SCANNED_DATE": types.DateTime(),
        "LAST_VM_AUTH_SCANNED_DURATION": types.Integer(),
        "LAST_VM_SCANNED_DURATION": types.Integer(),
        "LAST_VM_AUTH_SCANNED_DATE": types.DateTime(),
        "LAST_COMPLIANCE_SCAN_DATETIME": types.DateTime(),
        "LAST_PC_SCANNED_DATE": types.DateTime(),
        "ASSET_GROUP_IDS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "USER_DEF": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "OWNER": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "CLOUD_PROVIDER": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CLOUD_SERVICE": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CLOUD_RESOURCE_ID": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "EC2_INSTANCE_ID": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CLOUD_GROUP_NAME": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CLOUD_INSTANCE_STATE": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CLOUD_INSTANCE_TYPE": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CLOUD_IS_SPOT_INSTANCE": types.Boolean(),
        "CLOUD_ARCHITECTURE": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CLOUD_REGION": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CLOUD_IMAGE_ID": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CLOUD_AMI_ID": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CLOUD_PUBLIC_HOSTNAME": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CLOUD_PUBLIC_IPV4": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CLOUD_ACCOUNT_ID": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CLOUD_PROVIDER_TAGS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "ASSET_RISK_SCORE": types.Integer(),
        "TRURISK_SCORE": types.Integer(),
        "TRURISK_SCORE_FACTORS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "ASSET_CRITICALITY_SCORE": types.Integer(),
    }

    # Remove METADATA and DNS_DATA from the dataclass. They're parsed out already from dataclass initialization.
    df = DataFrame([prepare_dataclass(host) for host in hosts])

    df.drop(columns=["METADATA", "DNS_DATA", "DETECTION_LIST"], inplace=True)

    # Upload the data, with table depdening on if it is a Host List Detail pull or not:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_vmdr_ips(
    ips: BaseList,
    cnxn: Connection,
    table_name: str = "vmdr_ips",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from vmdr.get_ip_list() to SQL.

    Args:
        ips (BaseList): The IP List to upload from vmdr.get_ip_list().
        cnxn (Connection): The Connection object to the SQL database.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "IP": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "TYPE": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
    }

    # Convert the BaseList to a DataFrame:
    df = DataFrame([str(ip) for ip in ips], columns=["IP"])

    # Add the TYPE column, which shows if it is a single IP or a range:
    df["TYPE"] = df["IP"].apply(lambda x: "Single IP" if "/" not in x else "IP Range")

    # Upload the data:
    return upload_data(
        df, table_name, cnxn, dtype=COLS, override_import_dt=override_import_dt
    )


def upload_vmdr_hld(
    hld: BaseList,
    cnxn: Connection,
    vulns_table_name: str = "vmdr_hld_detections",
    hosts_table_name: str = "vmdr_hld_hosts_list",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from vmdr.get_hld() to SQL.

    Args:
        hld (BaseList): The Host List to upload.
        cnxn (Connection): The Connection object to the SQL database.
        vulns_table_name (str): The name of the table to upload the detections to. Defaults to 'vmdr_hld_detections'.
        hosts_table_name (str): The name of the table to upload the hosts to. Defaults to 'vmdr_hld_hosts_list'.

    Returns:
        int: The number of rows uploaded.

    """

    """
    Get_hld and get_host_list technically return the same data. get_hld just
    includes the DETECTION_LIST attribute. We can use the same upload function
    for the host part, and then snip off the DETECTION_LIST attribute to upload
    to a detections table.
    """

    # Isolate the detection lists. Since the Detection objects themselves
    # have an ID attribute, we can use that to link them back to the host.
    detections = BaseList()
    for host in hld:
        if host.DETECTION_LIST:
            for detection in host.DETECTION_LIST:
                detections.append(detection)

    # upload_vmdr_hosts automatically ignores the DETECTION_LIST attribute,
    # so we can use it here with the is_hld flag set to True to put the hosts
    # in a different table than get_host_list.
    hosts_uploaded = upload_vmdr_hosts(
        hld, cnxn, hosts_table_name, override_import_dt=override_import_dt
    )
    print(
        f"Uploaded {hosts_uploaded} hosts to {hosts_table_name}. Moving to detections..."
    )

    COLS = {
        "UNIQUE_VULN_ID": types.BigInteger(),
        "QID": types.BigInteger(),
        "TYPE": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "SEVERITY": types.Integer(),
        "STATUS": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "SSL": types.Boolean(),
        "RESULTS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "FIRST_FOUND_DATETIME": types.DateTime(),
        "LAST_FOUND_DATETIME": types.DateTime(),
        "QDS": types.Integer(),
        "QDS_FACTORS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "TIMES_FOUND": types.Integer(),
        "LAST_TEST_DATETIME": types.DateTime(),
        "LAST_UPDATE_DATETIME": types.DateTime(),
        "IS_IGNORED": types.Boolean(),
        "IS_DISABLED": types.Boolean(),
        "LAST_PROCESSED_DATETIME": types.DateTime(),
        "LAST_FIXED_DATETIME": types.DateTime(),
        "PORT": types.Integer(),
        "PROTOCOL": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "FQDN": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
    }

    # Create a copy of the detections list so we do not modify the original:
    detections = BaseList(detections)

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(detection) for detection in detections])

    # Set QDS to an integer:
    df["QDS"] = df["QDS"].apply(lambda x: int(x) if x else None)

    # Upload the data:
    return upload_data(
        df,
        vulns_table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_vmdr_scanners(
    scanners: BaseList,
    cnxn: Connection,
    table_name: str = "vmdr_scanners",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from vmdr.get_scanner_list() to SQL.

    Args:
        scanners (BaseList): The Scanner List to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_scanners'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "ID": types.Integer(),
        "NAME": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "SOFTWARE_VERSION": types.Float(),
        "RUNNING_SCAN_COUNT": types.Integer(),
        "STATUS": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "UUID": types.Uuid(),
        "RUNNING_SLICES_COUNT": types.Integer(),
        "MODEL_NUMBER": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "TYPE": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "SERIAL_NUMBER": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "ACTIVATION_CODE": types.BigInteger(),
        "INTERFACE_SETTINGS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "PROXY_SETTINGS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # dict.
        "IS_CLOUD_DEPLOYED": types.Boolean(),
        "CLOUD_INFO": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # dict. DROP THIS! INFO IS IN BELOW FIELDS:
        "PLATFORM_PROVIDER": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "INSTANCE_ID": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "AMI_ID": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "ACCOUNT_ID": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "INSTANCE_REGION": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "INSTANCE_AVAIBILITY_ZONE": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "INSTANCE_ZONE_TYPE": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "INSTANCE_VPC_ID": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "INSTANCE_SUBNET_ID": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "INSTANCE_ADDRESS_PRIVATE": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "INSTANCE_ADDRESS_PUBLIC": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "HOSTNAME_PRIVATE": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "SECURITY_GROUPS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "API_PROXY_SETTINGS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "VLANS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # dict.
        "STATIC_ROUTES": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # dict.
        "ML_LATEST": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "ML_VERSION": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "VULNSIGS_LATEST": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "VULNSIGS_VERSION": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "ASSET_GROUP_COUNT": types.Integer(),
        "ASSET_GROUP_LIST": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "LAST_UPDATED_DATE": types.DateTime(),
        "POLLING_INTERVAL": types.Integer(),
        "USER_LOGIN": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "HEARTBEATS_MISSED": types.Integer(),
        "SS_CONNECTIION": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "SS_LAST_CONNECTED": types.DateTime(),
        "USER_LIST": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # dict
        "UPDATED": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "COMMENTS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "MAX_CAPACITY_UNITS": types.Float(),
    }

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(scanner) for scanner in scanners])

    # Drop the CLOUD_INFO and EC2_INFO columns:
    df.drop(columns=["CLOUD_INFO"], inplace=True)

    # Upload the data:
    return upload_data(
        df, table_name, cnxn, dtype=COLS, override_import_dt=override_import_dt
    )


def upload_vmdr_static_search_lists(
    searchlists: BaseList,
    cnxn: Connection,
    table_name: str = "vmdr_static_searchlists",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from vmdr.get_static_searchlists() to SQL.

    Args:
        searchlists (BaseList): The Search List to upload.
        cnxn (Connection): The Connection object to the SQL database.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "ID": types.Integer(),
        "TITLE": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "GLOBAL": types.Boolean(),
        "OWNER": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "CREATED": types.DateTime(),
        "MODIFIED": types.DateTime(),
        "MODIFIED_BY": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "QIDS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "OPTION_PROFILES": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # dict
        "REPORT_TEMPLATES": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # dict
        "REMEDIATION_POLICIES": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # dict
        "DISTRIBUTION_GROUPS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # dict
        "COMMENTS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # dict
    }

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(searchlist) for searchlist in searchlists])

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_vmdr_dynamic_search_lists(
    searchlists: BaseList,
    cnxn: Connection,
    table_name: str = "vmdr_dynamic_searchlists",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from vmdr.get_dynamic_searchlists() to SQL.

    Args:
        searchlists (BaseList): The Search List to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_dynamic_searchlists'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "ID": types.Integer(),
        "TITLE": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "GLOBAL": types.Boolean(),
        "OWNER": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "CREATED": types.DateTime(),
        "MODIFIED": types.DateTime(),
        "MODIFIED_BY": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "QIDS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "CRITERIA": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "OPTION_PROFILES": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "REPORT_TEMPLATES": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "REMEDIATION_POLICIES": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "DISTRIBUTION_GROUPS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "COMMENTS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
    }

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(searchlist) for searchlist in searchlists])

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_vmdr_users(
    users: BaseList,
    cnxn: Connection,
    table_name: str = "vmdr_users",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from vmdr.get_user_list() to SQL.

    Args:
        users (BaseList): The user list to upload.
        cnxn (Connection): The Connection object to the SQL database.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "USER_LOGIN": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "USER_ID": types.Integer(),
        "EXTERNAL_ID": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "FIRSTNAME": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "LASTNAME": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "TITLE": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "PHONE": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "FAX": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "EMAIL": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "COMPANY": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "ADDRESS1": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "ADDRESS2": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CITY": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "COUNTRY": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "STATE": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "ZIP_CODE": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "TIME_ZONE_CODE": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "USER_STATUS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CREATION_DATE": types.DateTime(),
        "USER_ROLE": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "BUSINESS_UNIT": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "MANAGER_POC": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "UI_INTERFACE_STYLE": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "CREATE_OPTION_PROFILES": types.Boolean(),
        "PURGE_INFO": types.Boolean(),
        "ADD_ASSETS": types.Boolean(),
        "EDIT_REMEDIATION_POLICY": types.Boolean(),
        "EDIT_AUTH_RECORDS": types.Boolean(),
        "LATEST_VULN": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "MAP": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "DAILY_TICKETS": types.Boolean(),
        "ASSET_GROUP_TITLE": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "LAST_LOGIN_DATE": types.DateTime(),
        "UNIT_MANAGER_POC": types.Float(),
        "MAP": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "SCAN": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
    }

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(user) for user in users])

    # Drop contact_info, assigned_asset_groups,
    # permissions, notifications:
    df.drop(
        columns=[
            "CONTACT_INFO",
            "ASSIGNED_ASSET_GROUPS",
            "PERMISSIONS",
            "NOTIFICATIONS",
        ],
        inplace=True,
    )

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_vmdr_scan_list(
    scans: BaseList,
    cnxn: Connection,
    table_name: str = "vmdr_scans",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from vmdr.get_scan_list() to SQL.

    Args:
        scans (BaseList): The Scan List to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_scans'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "REF": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "TYPE": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "TITLE": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "USER_LOGIN": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "LAUNCH_DATETIME": types.DateTime(),
        "DURATION": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "PROCESSING_PRIORITY": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "PROCESSED": types.Boolean(),
        "STATE": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "TARGET": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "OPTION_PROFILE": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # dict
        "ASSET_GROUP_TITLE_LIST": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
    }

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(scan) for scan in scans])

    # Drop the STATUS column, as it is parsed out into STATE:
    df.drop(columns=["STATUS"], inplace=True)

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_vmdr_report_list(
    reports: BaseList,
    cnxn: Connection,
    table_name: str = "vmdr_reports",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from vmdr.get_report_list() to SQL.

    Args:
        reports (BaseList): The Report List of VMDRReports to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "ID": types.Integer(),
        "TITLE": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "TYPE": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "USER_LOGIN": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "LAUNCH_DATETIME": types.DateTime(),
        "OUTPUT_FORMAT": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "SIZE": types.Float(),
        "STATUS": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "EXPIRATION_DATETIME": types.DateTime(),
    }

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(report) for report in reports])

    # Drop the STATUS column, as it is parsed out into STATE:
    df.drop(columns=["STATUS"], inplace=True)

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_vmdr_scheduled_report_list(
    reports: BaseList,
    cnxn: Connection,
    table_name: str = "vmdr_scheduled_reports",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from vmdr.get_scheduled_report_list() to SQL.

    Args:
        reports (BaseList): The Report List of VMDRScheduledReports to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_scheduled_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "ID": types.Integer(),
        "TITLE": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "OUTPUT_FORMAT": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "TEMPLATE_TITLE": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "ACTIVE": types.Boolean(),
        "SCHEDULE": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # dict
        "START_DATE_UTC": types.DateTime(),
        "START_HOUR": types.Integer(),
        "START_MINUTE": types.Integer(),
        "TIME_ZONE_CODE": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "TIME_ZONE_DETAILS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "DST_SELECTED": types.Boolean(),
    }

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(report) for report in reports])

    # Drop TIME_ZONE column, as it is parsed out into TIME_ZONE_CODE and TIME_ZONE_DETAILS:
    df.drop(columns=["TIME_ZONE"], inplace=True)

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_vmdr_template_list(
    templates: BaseList,
    cnxn: Connection,
    table_name: str = "vmdr_templates",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from vmdr.get_template_list() to SQL.

    Args:
        templates (BaseList): The Report List of VMDRTemplates to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_templates'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "ID": types.Integer(),
        "TYPE": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "TEMPLATE_TYPE": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "TITLE": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "LOGIN": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "FIRSTNAME": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "LASTNAME": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "LAST_UPDATE": types.DateTime(),
        "GLOBAL": types.Boolean(),
    }

    def prepare_template(template) -> dict:
        """
        Helper function to use instead of prepare_dataclass
        to handle the USER attribute.

        df.drop is not necessary due to del data['USER'] here.

        Args:
            template (ReportTemplate): The template to prepare.

        Returns:
            dict: The template prepared for upload.
        """
        data = template.to_dict()
        data["LOGIN"] = template.USER.get("LOGIN")
        data["FIRSTNAME"] = template.USER.get("FIRSTNAME")
        data["LASTNAME"] = template.USER.get("LASTNAME")
        del data["USER"]

        return data

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_template(template) for template in templates])

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_vmdr_kb_qvs(
    qvs: BaseList,
    cnxn: Connection,
    table_name: str = "vmdr_kb_qvs",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from vmdr.get_kb_qvs() to SQL.

    Args:
        qvs (BaseList): The KBQVS List to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_kb_qvs'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "id": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "idType": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "qvs": types.Integer(),
        "qvsLastChangedDate": types.DateTime(),
        "nvdPublishedDate": types.DateTime(),
        "cvss": types.Float(),
        "cvssVersion": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "cvssString": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "epss": types.Float(),
        "threatActors": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "exploitMaturity": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "trending": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "mitigationControls": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "malwareName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "malwareHash": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "rti": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
    }

    df = DataFrame([prepare_dataclass(qv) for qv in qvs])

    df.drop(columns=["contributingFactors", "base"], inplace=True)

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_vmdr_activity_log(
    activity_log: BaseList,
    cnxn: Connection,
    table_name: str = "vmdr_activity_log",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from vmdr.get_activity_log() to SQL.

    Args:
        activity_log (BaseList): The Activity Log to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_activity_log'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "Date": types.DateTime(),
        "Action": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "Module": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "Details": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "User_Name": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "User_Role": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "User_IP": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
    }

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(log) for log in activity_log])

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )



"""
token.py - contains the TokenAuth class, which handles API endpoints that require JWT authentication
"""

from dataclasses import dataclass, field
from datetime import datetime

from requests import post

from .basic import BasicAuth
from ..exceptions import AuthenticationError


@dataclass
class TokenAuth(BasicAuth):
    """
    TokenAuth - handles API endpoints that require JWT authentication.
    This class will take the username, password, and platform attributes from BasicAuth and use them to generate a JWT token via the Qualys JWT-generatio API.

    Attributes:
    ```
    token: str - the JWT token for the API
    generated_on: datetime - the datetime the token was generated. Tokens are valid for 4 hours.
    ```
    Subclass of .basic.BasicAuth - provides the JWT authentication for the API
    """

    # JWT token
    token: str = field(init=False, repr=False)
    generated_on: datetime = field(init=False, default=datetime)

    def __post_init__(self):
        """
        generates the JWT token for the API
        """
        self.auth_type = "token"
        self.token = "placeholder"
        super().__post_init__()
        self.token = self.get_token()

    def get_token(self) -> str:
        """
        generates the JWT token from the Qualys API
        """
        url = f"https://gateway.{self.platform}.apps.qualys.com/auth"

        payload = {
            "username": self.username,
            "password": self.password,
            "permissions": "true",
            "token": "true",
        }

        headers = {"Content-Type": "application/x-www-form-urlencoded"}

        print(f"Generating token for {self.username} on {self.platform} platform.")

        r = post(url, headers=headers, data=payload)

        if r.status_code != 201:
            raise AuthenticationError(
                f"Failed to generate token. Requests reporting: {r.text}"
            )
        print("Success.")
        self.generated_on = datetime.now()
        return r.text

    def as_header(self) -> dict:
        """
        returns the headers for the API request
        """
        return {"Authorization": f"Bearer {self.token}"}

    @classmethod
    def from_dict(cls, data: dict):
        """
        provided the dictionary keys 'username', 'password', and 'platform' are present,
        creates a TokenAuth object from the dictionary
        """
        if all(key in data for key in ["username", "password"]):
            return cls(data["username"], data["password"])
        else:
            raise AuthenticationError(
                "Missing required keys 'username' or 'password' in dictionary."
            )



"""
Contains the functions to upload supported Container Security API pulls to SQL DBs.
"""

from datetime import datetime

from pandas import DataFrame
from sqlalchemy import Connection, types
from sqlalchemy.dialects.mysql import TEXT

from .base import upload_data, prepare_dataclass
from ..base.base_list import BaseList


def upload_was_webapps(
    webapps: BaseList,
    cnxn: Connection,
    table_name: str = "was_webapps",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload results from ```was.get_webapps```
    or ```was.get_webapps_verbose```
    to a SQL database.

    Args:
        webapps (BaseList): A BaseList of WebApp objects.
        cnxn (Connection): The Connection object to the SQL database.
        override_import_dt (datetime): If provided, will override the import_datetime column with this value.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "id": types.Integer(),
        "name": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "url": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "uris": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "riskScore": types.Integer(),
        "os": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "owner_id": types.Integer(),
        "owner_username": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "owner_firstName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "owner_lastName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "scope": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "attributes": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "defaultProfile_id": types.Integer(),
        "defaultProfile_name": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "defaultScanner_id": types.Integer(),
        "defaultScanner_name": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "defaultScannerTags_count": types.Integer(),
        "defaultScannerTags_list": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "scannerLocked": types.Boolean(),
        "progressiveScanning": types.Boolean(),
        "urlExcludelist": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "urlAllowlist": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "postDataExcludelist": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "logoutRegexList": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "authRecords": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "dnsOverrides": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "useRobots": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "useSitemap": types.Boolean(),
        "malwareMonitoring": types.Boolean(),
        "malwareNotifications": types.Boolean(),
        "tags": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "comments": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "isScheduled": types.Boolean(),
        "lastScan_id": types.Integer(),
        "lastScan_name": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "lastScan_summary_resultsStatus": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "lastScan_summary_authStatus": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "createdBy_id": types.Integer(),
        "createdBy_username": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "createdBy_firstName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "createdBy_lastName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "createdDate": types.DateTime(),
        "updatedDate": types.DateTime(),
        "updatedBy_id": types.Integer(),
        "updatedBy_username": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "updatedBy_firstName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "updatedBy_lastName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "screenshot": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "config": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "crawlingScripts": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "postmanCollection": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "headers": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "domains": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "subDomain": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "swaggerFile": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "redundancyLinks": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "maxRedundancyLinks": types.Integer(),
    }

    # Prepare the dataclass for insertion:
    df = DataFrame([prepare_dataclass(webapp) for webapp in webapps])

    # Drop any columns that we parsed out:
    df.drop(
        columns=[
            "owner",
            "defaultProfile",
            "defaultScanner",
            "defaultScannerTags",
            "lastScan",
            "createdBy",
            "updatedBy",
        ],
        inplace=True,
    )

    # Upload the data:
    return upload_data(df, table_name, cnxn, COLS, override_import_dt)



"""
Contains the WebAppAuthRecord class for the WAS module.
"""

from dataclasses import dataclass, asdict
from datetime import datetime
from typing import Union, Literal


@dataclass
class WebAppAuthRecord:
    """
    Represents an authentication record
    in Qualys WAS
    """

    id: Union[str, int] = None
    name: str = None
    owner: None = None
    # owner is parsed into below fields:
    owner_id: Union[str, int] = None
    owner_username: str = None
    owner_firstName: str = None
    owner_lastName: str = None
    # end owner
    tags: None = None
    # tags is parsed into below fields:
    tags_count: Union[str, int] = 0
    # end tags
    createdDate: Union[str, datetime] = None
    updatedDate: Union[str, datetime] = None

    def __post_init__(self):
        setattr(self, "id", int(self.id))

        DT_FIELDS = ["createdDate", "updatedDate"]

        for field in DT_FIELDS:
            if getattr(self, field) and not isinstance(getattr(self, field), datetime):
                setattr(self, field, datetime.fromisoformat(getattr(self, field)))

        if self.owner:
            self.owner_id = int(self.owner.get("id"))
            self.owner_username = self.owner.get("username")
            self.owner_firstName = self.owner.get("firstName")
            self.owner_lastName = self.owner.get("lastName")
            setattr(self, "owner", None)

        if self.tags:
            self.tags_count = len(self.tags)
            setattr(self, "tags", None)

    def to_dict(self):
        return asdict(self)

    def __dict__(self):
        return self.to_dict()

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)



"""
report.py - contains the VMDRReport dataclass for the Qualys VMDR module.
"""

from dataclasses import dataclass, field, asdict
from typing import Union, Dict
from datetime import datetime


@dataclass(order=True)
class VMDRReport:
    """
    Represents a VMDR report.
    """

    ID: int = field(
        metadata={"description": "The unique ID of the report."}, default=None
    )

    TITLE: str = field(
        metadata={"description": "The title of the report."}, default=None
    )

    TYPE: str = field(metadata={"description": "The type of the report."}, default=None)

    USER_LOGIN: str = field(
        metadata={"description": "The user that launched the report."}, default=None
    )

    LAUNCH_DATETIME: Union[str, datetime] = field(
        metadata={"description": "The date and time the report was launched."},
        default=None,
    )

    OUTPUT_FORMAT: str = field(
        metadata={"description": "The output format of the report."}, default=None
    )

    SIZE: Union[str, float] = field(
        metadata={"description": "The size of the report. Gets normalized to MB."},
        default=None,
    )

    STATUS: Union[Dict, str] = field(
        metadata={"description": "The status of the report."}, default=None
    )

    STATE: str = field(
        metadata={"description": "The state of the report."}, default=None
    )

    EXPIRATION_DATETIME: Union[str, datetime] = field(
        metadata={"description": "The date and time the report will expire."},
        default=None,
    )

    def __post_init__(self):
        DT_FIELDS = ["LAUNCH_DATETIME", "EXPIRATION_DATETIME"]
        for field in DT_FIELDS:
            if getattr(self, field):
                setattr(self, field, datetime.fromisoformat(getattr(self, field)))

        self.ID = int(self.ID)

        self.STATE = self.STATUS.get("STATE")

        if self.SIZE:
            # Now for the fun part. SIZE can be KB, MB, or maybe even GB. Let's convert it to bytes and then to MB.
            # incoming string will contain the unit, so we can just strip it off and convert to bytes.

            if self.SIZE == "-":
                self.SIZE = 0
                return

            # Split SIZE into the value and unit
            size_value, size_unit = self.SIZE.split(" ")
            size_value = float(size_value)
            size_unit = size_unit.upper()

            # Convert size to bytes
            match size_unit:
                case "MB":
                    # If we are already in MB, we can just return the value
                    self.SIZE = size_value
                    return

                case "KB":
                    size_in_bytes = size_value * 1024
                case "GB":
                    size_in_bytes = size_value * 1024**3
                case "BYTES":
                    size_in_bytes = size_value
                case _:
                    raise ValueError(f"Invalid unit {size_unit} for report size.")

            # Convert bytes to MB
            self.SIZE = size_in_bytes / 1024**2

    @classmethod
    def from_dict(cls, data: dict):
        """
        from_dict - creates a VMDRReport object from a dictionary.
        """
        return cls(**data)

    def __dict__(self):
        return asdict(self)

    def __int__(self):
        return self.ID

    def __str__(self):
        return f"{self.ID}: {self.TITLE}"

    def keys(self):
        return asdict(self).keys()

    def values(self):
        return asdict(self).values()

    def items(self):
        return asdict(self).items()

    def to_dict(self):
        return asdict(self)

    def valid_values(self):
        """
        valid_values - returns a dictionary of attributes that have non-None values.
        """
        return {k: v for k, v in self.to_dict().items() if v}


@dataclass(order=True)
class VMDRScheduledReport:
    """
    Represents a VMDR scheduled report.
    """

    ID: int = field(
        metadata={"description": "The unique ID of the report."}, default=None
    )

    TITLE: str = field(
        metadata={"description": "The title of the report."}, default=None
    )

    OUTPUT_FORMAT: str = field(
        metadata={"description": "The output format of the report."}, default=None
    )

    TEMPLATE_TITLE: str = field(
        metadata={"description": "The title of the report template."}, default=None
    )

    ACTIVE: Union[bool, str] = field(
        metadata={"description": "Whether the report is active."}, default=None
    )

    SCHEDULE: dict = field(
        metadata={"description": "The schedule of the report."}, default=None
    )

    START_DATE_UTC: Union[str, datetime] = field(
        metadata={"description": "The start date of the report."}, default=None
    )

    START_HOUR: int = field(
        metadata={"description": "The start hour of the report."}, default=None
    )

    START_MINUTE: int = field(
        metadata={"description": "The start minute of the report."}, default=None
    )

    TIME_ZONE: dict = field(
        metadata={"description": "The time zone of the report."}, default=None
    )

    TIME_ZONE_CODE: str = field(
        metadata={"description": "The time zone code of the report."}, default=None
    )

    TIME_ZONE_DETAILS: str = field(
        metadata={"description": "The time zone details of the report."}, default=None
    )

    DST_SELECTED: Union[bool, str] = field(
        metadata={"description": "Whether DST is selected."}, default=None
    )

    def __post_init__(self):
        self.ID = int(self.ID)

        if self.ACTIVE:
            self.ACTIVE = bool(self.ACTIVE)

        if self.SCHEDULE:
            # First, do all the conversions.
            self.START_DATE_UTC = datetime.fromisoformat(
                self.SCHEDULE["START_DATE_UTC"]
            )
            self.START_HOUR = int(self.SCHEDULE["START_HOUR"])
            self.START_MINUTE = int(self.SCHEDULE["START_MINUTE"])
            self.TIME_ZONE = self.SCHEDULE["TIME_ZONE"]
            self.TIME_ZONE_CODE = self.TIME_ZONE["TIME_ZONE_CODE"]
            self.TIME_ZONE_DETAILS = self.TIME_ZONE["TIME_ZONE_DETAILS"]
            self.DST_SELECTED = bool(self.SCHEDULE["DST_SELECTED"])
            # Finally, remove all the above keys from the SCHEDULE dict,
            # leaving us only with the frequency data.
            for key in [
                "START_DATE_UTC",
                "START_HOUR",
                "START_MINUTE",
                "TIME_ZONE",
                "DST_SELECTED",
            ]:
                self.SCHEDULE.pop(key)

    @classmethod
    def from_dict(cls, data: dict):
        """
        from_dict - creates a VMDRScheduledReport object from a dictionary.
        """
        return cls(**data)

    def __dict__(self):
        return asdict(self)

    def __int__(self):
        return self.ID

    def __str__(self):
        return f"{self.ID}: {self.TITLE}"

    def keys(self):
        return asdict(self).keys()

    def values(self):
        return asdict(self).values()

    def items(self):
        return asdict(self).items()

    def valid_values(self):
        """
        valid_values - returns a dictionary of attributes that have non-None values.
        """
        return {k: v for k, v in self.to_dict().items() if v}

    def to_dict(self):
        return asdict(self)



"""
get_all_assets.py - contains the get_all_assets function for the Global AssetView API (GAV) module.
"""

from typing import Union

from ..base.base_list import BaseList
from ..base.call_api import call_api
from ..auth.token import TokenAuth
from ..exceptions.Exceptions import *
from .hosts import Host


def get_all_assets(
    auth: TokenAuth, page_count: Union[int, "all"] = "all", **kwargs
) -> BaseList[Host]:
    """
    Get all assets in the Global AssetView API.

    Params:
        auth (TokenAuth): The authentication object.
        page_count (Union[int, "all"]): The number of pages to get. If "all", get all pages. Defaults to "all".

    :Kwargs:
        excludeFields (str): The fields to exclude.
        includeFields (str): The fields to include.
        lastSeenAssetId (int): The last seen asset ID. Used for automatic pagination.
        lastModifiedDate (str): The last modified date.
        pageSize (int): The number of assets to get per page.

    Returns:
        BaseList[Hosts]: The response from the API as a BaseList of Hosts objects.
    """

    responses = BaseList()
    pulled = 0

    while True:
        # make the request:
        response = call_api(
            auth=auth, module="gav", endpoint="get_all_assets", params=kwargs
        )
        j = response.json()
        for record in j["assetListData"]["asset"]:
            responses.append(Host(**record))
        (
            print(f"Page {pulled+1} of {page_count} complete.")
            if page_count != "all"
            else print(f"Page {pulled+1} complete.")
        )
        pulled += 1

        if not j["hasMore"]:
            print("No more records.")
            break

        if page_count != "all" and pulled >= page_count:
            print("Page count reached.")
            break

        else:
            kwargs["lastSeenAssetId"] = j["lastSeenAssetId"]

    print("All pages complete.")
    return responses



"""
Contains user-facing functions for interacting with authentication records in WAS
"""

from typing import Union

from .data_classes.WebAppAuthRecord import WebAppAuthRecord
from .base.service_requests import build_service_request, build_update_request
from .base.parse_kwargs import validate_kwargs
from .base.service_requests import validate_response
from ..base.call_api import call_api
from ..auth.basic import BasicAuth
from ..exceptions.Exceptions import QualysAPIError
from ..base.base_list import BaseList


def call_auth_api(
    auth: BasicAuth, endpoint: str, payload: dict
) -> Union[int, WebAppAuthRecord]:
    """
    Call a Qualys WAS API auth record endpoint and return the parsed response. This is
    a backend function and should not be called directly.

    Args:
        auth (BasicAuth): The authentication object.
        endpoint (str): The API endpoint to call.
        payload (dict): The payload to send to the API.

    Returns:
        Union[int, WebAppAuthRecord]: The parsed response from the API.
    """

    match endpoint:
        case "count_authentication_records":
            params = {"placeholder": "count", "webappAuthRecordId": ""}
        case "get_authentication_records":
            params = {"placeholder": "search", "webappAuthRecordId": ""}
        case _:
            raise ValueError(f"Invalid endpoint: {endpoint}")

    response = call_api(
        auth=auth,
        override_method=(
            "GET" if endpoint == "get_authentication_record_details" else "POST"
        ),
        module="was",
        endpoint="call_auth_api",
        params=params,
        payload=payload,
        headers={"Content-Type": "text/xml"},
    )

    return validate_response(response)


def count_authentication_records(auth: BasicAuth, **kwargs) -> int:
    """
    Return a count of auth records in the Qualys WAS module
    according to the provided filters.

    Args:
        auth (BasicAuth): The authentication object.

    ## Kwargs:

        - id (Union[str, int]): Authentication object ID.
        - id_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the ID filter.
        - name (str): Auth object name.
        - name_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the name filter.
        - tags (Union[str, int]): Tag ID.
        - tags_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the tags filter.
        - tags_name (str): Tag name.
        - tags_name_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the tag name filter.
        - tags_id (Union[str, int]): Tag ID.
        - tags_id_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the tag ID filter.
        - createdDate (str): Date the Auth object was created in UTC date/time format.
        - createdDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the created date filter.
        - updatedDate (str): Date the Auth object was last updated in UTC date/time format.
        - updatedDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the updated date filter.
        - isUsed (bool): Whether the Auth object is in use.
        - isUsed (Literal["EQUALS", "NOT EQUALS"]): Operator for the isUsed filter.
        - lastScan_authStatus (Literal["NONE", "NOT_USED", "PARTIAL", "FAILED", "SUCCESSFUL"]): Status of the last scan.
        - lastScan_authStatus_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the last scan status filter.
        - lastScan_date (str): Date of the last scan in UTC date/time format.
        - lastScan_date_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the last scan date filter.
        - contents (Literal["FORM_STANDARD", "FORM_CUSTOM", "FORM_SELENIUM", "SERVER_BASIC", "SERVER_DIGEST", "SERVER_NTLM", "CERTIFICATE", "OAUTH2_AUTH_CODE", "OAUTH2_IMPLICIT", "OAUTH2_PASSWORD", "OAUTH2_CLIENT_CREDS"]): Auth object contents.
        - contents_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the contents filter.

    Returns:
        int: The count of Auth objects that match the filters.
    """

    payload = None

    # If kwargs are provided, validate them:
    if kwargs:
        kwargs = validate_kwargs(endpoint="count_authentication_records", **kwargs)
        payload = build_service_request(**kwargs)

    # Make the API call:
    parsed = call_auth_api(auth, "count_authentication_records", payload)

    serviceResponse = parsed.get("ServiceResponse")
    if not serviceResponse:
        raise QualysAPIError("No ServiceResponse tag returned in the API response")

    if serviceResponse.get("responseCode") != "SUCCESS":
        raise QualysAPIError(
            f"API response returned error: {serviceResponse.get('responseCode')}"
        )

    return int(serviceResponse.get("count"))


def get_authentication_records(
    auth: BasicAuth, page_count: Union[int, "all"] = "all", **kwargs
) -> object:
    """
    Get a list of auth records in the Qualys WAS module
    according to the provided filters.

    Args:
        auth (BasicAuth): The authentication object.
        page_count (int): The number of pages to return. If 'all', return all pages. Default is 'all'.

    ## Kwargs:

        - id (Union[str, int]): Authentication object ID.
        - id_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the ID filter.
        - name (str): Auth object name.
        - name_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the name filter.
        - tags (Union[str, int]): Tag ID.
        - tags_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the tags filter.
        - tags_name (str): Tag name.
        - tags_name_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the tag name filter.
        - tags_id (Union[str, int]): Tag ID.
        - tags_id_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the tag ID filter.
        - createdDate (str): Date the Auth object was created in UTC date/time format.
        - createdDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the created date filter.
        - updatedDate (str): Date the Auth object was last updated in UTC date/time format.
        - updatedDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the updated date filter.
        - isUsed (bool): Whether the Auth object is in use.
        - isUsed (Literal["EQUALS", "NOT EQUALS"]): Operator for the isUsed filter.
        - lastScan_authStatus (Literal["NONE", "NOT_USED", "PARTIAL", "FAILED", "SUCCESSFUL"]): Status of the last scan.
        - lastScan_authStatus_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the last scan status filter.
        - lastScan_date (str): Date of the last scan in UTC date/time format.
        - lastScan_date_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the last scan date filter.
        - contents (Literal["FORM_STANDARD", "FORM_CUSTOM", "FORM_SELENIUM", "SERVER_BASIC", "SERVER_DIGEST", "SERVER_NTLM", "CERTIFICATE", "OAUTH2_AUTH_CODE", "OAUTH2_IMPLICIT", "OAUTH2_PASSWORD", "OAUTH2_CLIENT_CREDS"]): Auth object contents.
        - contents_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the contents filter.

    Returns:
        object: The auth records that match the filters.
    """

    # Ensure that page_count is either the string 'all' or an integer >= 1:
    if isinstance(page_count, int) and page_count < 1:
        raise QualysAPIError("page_count must be 'all' or an integer >= 1.")
    elif isinstance(page_count, str) and page_count != "all":
        raise QualysAPIError("page_count must be 'all' or an integer >= 1.")

    pageNo = 0
    payload = None

    # If kwargs are provided, validate them:
    if kwargs:
        kwargs = validate_kwargs(endpoint="get_authentication_records", **kwargs)
        payload = build_service_request(**kwargs)

    appList = BaseList()

    while True:
        # Make the API call:
        parsed = call_auth_api(auth, "get_authentication_records", payload)

        # Parse the XML response:
        serviceResponse = parsed.get("ServiceResponse")
        if not serviceResponse:
            raise QualysAPIError("No ServiceResponse tag returned in the API response")

        if serviceResponse.get("responseCode") != "SUCCESS":
            raise QualysAPIError(
                f"API response returned error: {serviceResponse.get('responseCode')}"
            )

        if serviceResponse.get("count") == "0":
            print(f"No web applications found on page {pageNo}. Exiting.")
            break

        data = serviceResponse.get("data")

        if data.get("WebAppAuthRecord"):
            data = data.get("WebAppAuthRecord")

        if isinstance(data, dict):
            data = [data]

        for record in data:
            # Create the objects:
            appList.append(WebAppAuthRecord.from_dict(record))

        print(
            f"Retrieved {serviceResponse.get('count')} auth records on page {pageNo}. Running total: {len(appList)}"
        )

        pageNo += 1

        if page_count != "all" and pageNo >= page_count:
            print(f"Reached page_count limit. Returning {pageNo} page(s).")
            break

        # Check for pagination:
        if serviceResponse.get("hasMoreRecords") == "true":
            # Update the XML payload with the new Criteria:
            # <Criteria field="id" operator="GREATER">XXX</Criteria>
            kwargs["id.operator"] = "GREATER"
            kwargs["id"] = serviceResponse.get("lastId")
            payload = build_service_request(**kwargs)
        else:
            break

    return appList



"""
Contains the Comment class for WAS
"""

from dataclasses import dataclass, asdict
from typing import Union
from datetime import datetime


@dataclass
class Comment:
    """
    Represents a comment on a WAS object
    """

    contents: str = None
    author_id: Union[str, int] = None
    author_username: str = None
    createdDate: Union[str, datetime] = None

    def __post_init__(self):
        if self.createdDate and isinstance(self.createdDate, str):
            # format: 2023-12-08T09:57:51Z
            setattr(
                self,
                "createdDate",
                datetime.strptime(self.createdDate, "%Y-%m-%dT%H:%M:%SZ"),
            )

        if self.author_id:
            try:
                self.author_id = int(self.author_id)
            except ValueError:
                raise ValueError("author_id must be numeric")

    def __str__(self) -> str:
        return self.contents

    def to_dict(self):
        """
        Converts the object to a dictionary
        """
        return asdict(self)

    def __dict__(self):
        return self.to_dict()

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @classmethod
    def from_dict(cls, data):
        return cls(**data)



"""
cve.py - contains the CVEID dataclass for the Qualys VMDR module.
"""

from dataclasses import dataclass, field, asdict
from typing import *
from re import match


@dataclass
class CVEID:
    """
    CVEID - represents a single CVE ID in a CVEList.
    """

    ID: str = field(metadata={"description": "The ID of the CVE."}, compare=True)
    URL: Optional[str] = field(
        metadata={"description": "The URL of the CVE."}, default="", compare=False
    )

    def __post_init__(self):
        # make sure that the ID is a string:
        if not isinstance(self.ID, str):
            raise TypeError(f"CVEID ID must be a string, not {type(self.ID)}")
        # and that url is a string:
        if not isinstance(self.URL, str):
            raise TypeError(f"CVEID URL must be a string, not {type(self.URL)}")

    def __str__(self) -> str:
        return self.ID

    def __dict__(self):
        return asdict(self)

    def __contains__(self, item):
        # see if it was found in the name or vendor:
        return item in self.ID

    def copy(self):
        return CVEID(id=self.ID)

    def is_id(self, id: str):
        return self.ID == id

    def to_dict(self):
        return asdict(self)

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @classmethod
    def from_str(cls, cve_id: str):
        """
        from_str - create a CVEID object from a string.

        This function is used to create a CVEID object from a string.

        The string should be in the format "CVE-YYYY-NNNN".

        If the string is not in the correct format, a ValueError will be raised.
        """
        cve_regex = r"CVE-\d{4}-\d{4,}"
        if not match(cve_regex, cve_id):
            raise ValueError(f"Invalid CVE ID format: {cve_id}")

        return cls(ID=cve_id)

    @classmethod
    def from_dict(cls, data: dict):
        """
        from_dict - create a CVEID object from a dictionary.

        This function is used to create a CVEID object from a dictionary.
        """
        # make sure that the dictionary has the required keys and nothing else:
        required_keys = {"ID"}
        if not required_keys.issubset(data.keys()):
            raise ValueError(
                f"Dictionary must contain the following keys: {required_keys}"
            )

        return cls(**data)



"""
Contains functions to interact with Web applications in the Qualys WAS module.
"""

from typing import Union

from .data_classes.WebApp import WebApp
from .base.service_requests import build_service_request, build_update_request
from .base.parse_kwargs import validate_kwargs
from .base.service_requests import validate_response
from ..base.call_api import call_api
from ..auth.basic import BasicAuth
from ..exceptions.Exceptions import QualysAPIError
from ..base.base_list import BaseList


def call_webapp_api(
    auth: BasicAuth, endpoint: str, payload: dict
) -> Union[int, WebApp]:
    """
    Call a Qualys WAS API endpoint and return the parsed response. This is
    a backend function and should not be called directly.

    Args:
        auth (BasicAuth): The authentication object.
        endpoint (str): The API endpoint to call.
        payload (dict): The payload to send to the API.

    Returns:
        Union[int, WebApp]: The parsed response from the API.
    """

    match endpoint:
        case "count_webapps":
            params = {"placeholder": "count", "webappId": ""}
        case "get_webapps":
            params = {"placeholder": "search", "webappId": ""}
        case "get_webapp_details":
            params = {"placeholder": "get", "webappId": payload.pop("webappId")}
        case "create_webapp":
            params = {"placeholder": "create", "webappId": ""}
        case "update_webapp":
            params = {"placeholder": "update", "webappId": payload.pop("webappId")}
        case "delete_webapp":
            params = {"placeholder": "delete", "webappId": ""}
            if payload.pop("removeFromSubscription", None):
                params["action"] = "removeFromSubscription"
        case "get_selenium_script":
            params = {"placeholder": "downloadSeleniumScript", "webappId": ""}
        case _:
            raise ValueError(f"Invalid endpoint: {endpoint}")

    response = call_api(
        auth=auth,
        override_method="GET" if endpoint == "get_webapp_details" else "POST",
        module="was",
        endpoint="call_webapp_api",
        params=params,
        payload=payload,
        headers={"Content-Type": "text/xml"},
    )

    return validate_response(response)


def count_webapps(auth: BasicAuth, **kwargs) -> int:
    """
    Return a count of web applications in the Qualys WAS module
    according to the provided filters.

    Args:
        auth (BasicAuth): The authentication object.

    ## Kwargs:

        - id (Union[str, int]): Web application ID.
        - id_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the ID filter.
        - name (str): Web application name.
        - name_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the name filter.
        - url (str): Web application URL.
        - url_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the URL filter.
        - tags_name (str): Tag name.
        - tags_name_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the tag name filter.
        - tags_id (Union[str, int]): Tag ID.
        - tags_id_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the tag ID filter.
        - createdDate (str): Date the web application was created in UTC date/time format.
        - createdDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the created date filter.
        - updatedDate (str): Date the web application was last updated in UTC date/time format.
        - updatedDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the updated date filter.
        - isScheduled (bool): Whether the web application has a scan scheduled.
        - isScheduled_operator (Literal["EQUALS", "NOT EQUALS"]): Operator for the isScheduled filter.
        - isScanned (bool): Whether the web application has been scanned.
        - isScanned_operator (Literal["EQUALS", "NOT EQUALS"]): Operator for the isScanned filter.
        - lastScan_status (Literal["SUBMITTED", "RUNNING", "FINISHED", "TIME_LIMIT_EXCEEDED", "SCAN_NOT_LAUNCHED", "SCANNER_NOT_AVAILABLE", "ERROR", "CANCELLED"]): Status of the last scan.
        - lastScan_status_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the last scan status filter.
        - lastScan_date (str): Date of the last scan in UTC date/time format.
        - lastScan_date_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the last scan date filter.

    Returns:
        int: The count of web applications that match the filters.
    """

    payload = None

    # If kwargs are provided, validate them:
    if kwargs:
        # kwargs = validate_kwargs(endpoint="count_webapps", **kwargs)
        kwargs = validate_kwargs(endpoint="count_webapps", **kwargs)
        payload = build_service_request(**kwargs)

    # Make the API call:
    parsed = call_webapp_api(auth, "count_webapps", payload)

    serviceResponse = parsed.get("ServiceResponse")
    if not serviceResponse:
        raise QualysAPIError("No ServiceResponse tag returned in the API response")

    if serviceResponse.get("responseCode") != "SUCCESS":
        raise QualysAPIError(
            f"API response returned error: {serviceResponse.get('responseCode')}"
        )

    return int(serviceResponse.get("count"))


def get_webapps(
    auth: BasicAuth, page_count: Union[int, "all"] = "all", **kwargs
) -> WebApp:
    """
    Get a list of web applications in the Qualys WAS module
    according to the provided filters.

    Args:
        auth (BasicAuth): The authentication object.
        page_count (int): The number of pages to return. If 'all', return all pages. Default is 'all'.

    ## Kwargs:

        - id (Union[str, int]): Web application ID.
        - id_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the ID filter.
        - name (str): Web application name.
        - name_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the name filter.
        - url (str): Web application URL.
        - url_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the URL filter.
        - tags_name (str): Tag name.
        - tags_name_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the tag name filter.
        - tags_id (Union[str, int]): Tag ID.
        - tags_id_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the tag ID filter.
        - createdDate (str): Date the web application was created in UTC date/time format.
        - createdDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the created date filter.
        - updatedDate (str): Date the web application was last updated in UTC date/time format.
        - updatedDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the updated date filter.
        - isScheduled (bool): Whether the web application has a scan scheduled.
        - isScheduled_operator (Literal["EQUALS", "NOT EQUALS"]): Operator for the isScheduled filter.
        - isScanned (bool): Whether the web application has been scanned.
        - isScanned_operator (Literal["EQUALS", "NOT EQUALS"]): Operator for the isScanned filter.
        - lastScan_status (Literal["SUBMITTED", "RUNNING", "FINISHED", "TIME_LIMIT_EXCEEDED", "SCAN_NOT_LAUNCHED", "SCANNER_NOT_AVAILABLE", "ERROR", "CANCELLED"]): Status of the last scan.
        - lastScan_status_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the last scan status filter.
        - lastScan_date (str): Date of the last scan in UTC date/time format.
        - lastScan_date_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the last scan date filter.
        - verbose (bool): If True, returns tag information in the response.

    Returns:
        WebApp: The web applications that match the filters.
    """

    # Ensure that page_count is either the string 'all' or an integer >= 1:
    if isinstance(page_count, int) and page_count < 1:
        raise QualysAPIError("page_count must be 'all' or an integer >= 1.")
    elif isinstance(page_count, str) and page_count != "all":
        raise QualysAPIError("page_count must be 'all' or an integer >= 1.")

    pageNo = 0
    payload = None

    # If kwargs are provided, validate them:
    if kwargs:
        kwargs = validate_kwargs(endpoint="get_webapps", **kwargs)
        payload = build_service_request(**kwargs)

    appList = BaseList()

    while True:
        # Make the API call:
        parsed = call_webapp_api(auth, "get_webapps", payload)

        # Parse the XML response:
        serviceResponse = parsed.get("ServiceResponse")
        if not serviceResponse:
            raise QualysAPIError("No ServiceResponse tag returned in the API response")

        if serviceResponse.get("responseCode") != "SUCCESS":
            raise QualysAPIError(
                f"API response returned error: {serviceResponse.get('responseCode')}"
            )

        if serviceResponse.get("count") == "0":
            print(f"No web applications found on page {pageNo}. Exiting.")
            break

        data = serviceResponse.get("data")

        if data.get("WebApp"):
            data = data.get("WebApp")

        if isinstance(data, dict):
            data = [data]

        for webapp in data:
            # Create the objects:
            appList.append(WebApp.from_dict(webapp))

        print(
            f"Retrieved {serviceResponse.get('count')} web applications on page {pageNo}. Running total: {len(appList)}"
        )

        pageNo += 1

        if page_count != "all" and pageNo >= page_count:
            print(f"Reached page_count limit. Returning {pageNo} page(s).")
            break

        # Check for pagination:
        if serviceResponse.get("hasMoreRecords") == "true":
            # Update the XML payload with the new Criteria:
            # <Criteria field="id" operator="GREATER">XXX</Criteria>
            kwargs["id.operator"] = "GREATER"
            kwargs["id"] = serviceResponse.get("lastId")
            payload = build_service_request(**kwargs)
        else:
            break

    return appList


def get_webapp_details(
    auth: BasicAuth, webappId: Union[int, str]
) -> Union[WebApp, None]:
    """
    Get the details of a single web application in the Qualys WAS module.

    Args:
        auth (BasicAuth): The authentication object.
        webappId (Union[int, str]): The ID of the web application.

    Returns:
        WebApp: The web application details.
    """

    # Make the API call:
    parsed = call_webapp_api(auth, "get_webapp_details", {"webappId": webappId})

    serviceResponse = parsed.get("ServiceResponse")
    if not serviceResponse:
        raise QualysAPIError("No ServiceResponse tag returned in the API response")

    if serviceResponse.get("responseCode") != "SUCCESS":
        raise QualysAPIError(
            f"API response returned error: {serviceResponse.get('responseCode')}"
        )

    data = serviceResponse.get("data")
    if data.get("WebApp"):
        data = data.get("WebApp")
        return WebApp.from_dict(data)
    else:
        print(f"No data found for web application ID {webappId}. Exiting.")


def get_webapps_verbose(
    auth: BasicAuth, thread_count: int = 5, **kwargs
) -> BaseList[WebApp]:
    """
    Uses ```was.get_webapps()``` and ```was.get_webapp_details()``` to return a ```BaseList``` of ```WebApp```s with
    all attributes populated.

    This function is multi-threaded, placing all ```WebApps``` found
    from ```was.get_webapps()``` into a queue and then spawning threads to pull the details one
    by one.

    The details threads wait for work to be added to the queue and then pull
    the details for each ```WebApp.id```.

    Args:
        auth (BasicAuth): The authentication object.
        thread_count (int): The number of threads to spawn. Defaults to 5.

    ## Kwargs:

        - id (Union[str, int]): Web application ID.
        - id_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the ID filter.
        - name (str): Web application name.
        - name_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the name filter.
        - url (str): Web application URL.
        - url_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the URL filter.
        - tags_name (str): Tag name.
        - tags_name_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the tag name filter.
        - tags_id (Union[str, int]): Tag ID.
        - tags_id_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the tag ID filter.
        - createdDate (str): Date the web application was created in UTC date/time format.
        - createdDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the created date filter.
        - updatedDate (str): Date the web application was last updated in UTC date/time format.
        - updatedDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the updated date filter.
        - isScheduled (bool): Whether the web application has a scan scheduled.
        - isScheduled_operator (Literal["EQUALS", "NOT EQUALS"]): Operator for the isScheduled filter.
        - isScanned (bool): Whether the web application has been scanned.
        - isScanned_operator (Literal["EQUALS", "NOT EQUALS"]): Operator for the isScanned filter.
        - lastScan_status (Literal["SUBMITTED", "RUNNING", "FINISHED", "TIME_LIMIT_EXCEEDED", "SCAN_NOT_LAUNCHED", "SCANNER_NOT_AVAILABLE", "ERROR", "CANCELLED"]): Status of the last scan.
        - lastScan_status_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the last scan status filter.
        - lastScan_date (str): Date of the last scan in UTC date/time format.
        - lastScan_date_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the last scan date filter.

    Returns:
        BaseList[WebApp]: The list of web applications.
    """

    # Check thread_count for a valid value:
    if not isinstance(thread_count, int) or thread_count < 1:
        raise ValueError("thread_count must be an integer >= 1.")

    # Set up Queue and List, with some cheeky as-needed imports:
    from queue import Queue
    from threading import Thread, Lock, current_thread

    q = Queue()
    appList = BaseList()
    LOCK = Lock()
    threads = []

    # Get the webapps:
    print(f"({current_thread().name}) Getting base Webapp list...")
    webapps = get_webapps(auth, page_count="all", **kwargs)

    print(
        f"({current_thread().name}) Pulled {len(webapps)} webapps. Starting {thread_count} thread(s) for details pull.."
    )

    # Add the webapps to the queue:
    for webapp in webapps:
        q.put(webapp)

    def worker():
        while True:
            try:
                # Exit condition 1: Queue is empty
                if q.empty():
                    with LOCK:
                        print(
                            f"({current_thread().name}) Queue is empty. Thread exiting."
                        )
                        break

                webapp = q.get()
                # Exit condition 2: webapp is None (because Queue is empty)
                if not webapp:
                    with LOCK:
                        print(
                            f"({current_thread().name}) Queue is empty. Thread exiting."
                        )
                        q.task_done()
                    break

                details = get_webapp_details(auth, webapp.id)
                appList.append(details)
                q.task_done()
                with LOCK:
                    if len(appList) % 10 == 0:
                        print(
                            f"({current_thread().name}) Pulled {len(appList)} webapp details so far..."
                        )

            except Exception as e:
                with LOCK:
                    print(
                        f"[ERROR - THREAD EXITING] ({current_thread().name}) Error: {e}"
                    )
                q.task_done()
                break

    # Start the threads:
    for i in range(thread_count):
        t = Thread(target=worker)
        threads.append(t)
        t.start()

    # Wait for the threads to finish:
    for t in threads:
        t.join()

    print(f"Pulled {len(appList)} webapp details.")
    return appList


def create_webapp(auth: BasicAuth, name: str, url: str, **kwargs) -> WebApp:
    """
    Create a new WAS web application.

    Args:
        auth (BasicAuth): The authentication object.
        name (str): The name of the web application.
        url (str): The URL of the web application.

    ## Kwargs:

            - authRecord_id (Union[str, int]): A single authentication record ID to associate with the web application.
            - uris (Union[str, list[str]]): A list or comma-separated string of URIs to associate with the web application.
            - tag_ids (Union[int, list[int]]): A single tag ID or a list of tag IDs to associate with the web application.
            - domains (Union[str, list[str]]): A single domain or a list of domains to associate with the web application.
            - scannerTag_ids (Union[int, list[int]]): A tag ID representing 1+ scanners to associate with the web application.

    Returns:
        WebApp: The new web application as a qualysdk WAS WebApp object.
    """

    kwargs["name"] = name
    kwargs["url"] = url

    # Validate the kwargs:
    kwargs = validate_kwargs(endpoint="create_webapp", **kwargs)

    # Build the XML payload:
    payload = build_service_request(
        _webapp_creation_or_edit=True,
        authRecord_id=kwargs.get("authRecord.id"),
        _uris=kwargs.get("uris"),
        tag_ids=kwargs.get("tag.ids"),
        _domains=kwargs.get("domains"),
        _scannerTag_ids=kwargs.get("scannerTag.ids"),
        **kwargs,
    )

    # Make the API call:
    parsed = call_webapp_api(auth, "create_webapp", payload)

    serviceResponse = parsed.get("ServiceResponse")
    if not serviceResponse:
        raise QualysAPIError("No ServiceResponse tag returned in the API response")

    if serviceResponse.get("responseCode") != "SUCCESS":
        raise QualysAPIError(
            f"API response returned error: {serviceResponse.get('responseCode')}"
        )

    data = serviceResponse.get("data")
    if data.get("WebApp"):
        data = data.get("WebApp")
        return WebApp.from_dict(data)
    else:
        print("No data found. Exiting.")


def update_webapp(auth: BasicAuth, webappId: Union[int, str], **kwargs) -> str:
    """
    Update a WAS web application.

    Args:
        auth (BasicAuth): The authentication object.
        webappId (Union[int, str]): The ID of the web application.

    ## Kwargs:

            - name (str): The name of the web application.
            - url (str): The URL of the web application.
            - attributes (dict["add": {"attr_name": "attr_value"}, "remove": ["attr_name"]]): Dictionary containing a list of attributes to add or remove.)
            - defaultProfile_id (Union[str, int]): The ID of the default profile to associate with the web application.
            - urlExcludelist (list[str]): A list of URLs to exclude from the scan.
            - urlAllowlist (list[str]): A list of URLs to allow in the scan.
            - postDataExcludelist (list[str]): A list of post data to exclude from the scan.
            - useSitemap (bool): Whether to use the sitemap.
            - headers (list[str]): A list of headers to include in the scan.
            - authRecord_id (dict["add": [int], "remove": [int]]): A dictionary containing a list of authentication record IDs to add or remove.

    Returns:
        str: A status message indicating the result of the update.
    """

    # Validate the kwargs:
    kwargs = validate_kwargs(endpoint="update_webapp", **kwargs)

    # Build the XML payload:
    payload = build_update_request(
        _authRecord_id=kwargs.get("authRecord.id"),
        tag_ids=kwargs.get("tag.ids"),
        _domains=kwargs.get("domains"),
        _scannerTag_ids=kwargs.get("scannerTag.ids"),
        defaultProfile_id=kwargs.get("defaultProfile.id"),
        _urlExcludelist=kwargs.get("urlExcludelist"),
        _urlAllowlist=kwargs.get("urlAllowlist"),
        _postDataExcludelist=kwargs.get("postDataExcludelist"),
        _useSitemap=kwargs.get("useSitemap"),
        _headers=kwargs.get("headers"),
        **kwargs,
    )

    # Make the API call:
    payload["webappId"] = webappId
    parsed = call_webapp_api(auth, "update_webapp", payload)

    serviceResponse = parsed.get("ServiceResponse")
    if not serviceResponse:
        raise QualysAPIError("No ServiceResponse tag returned in the API response")

    if serviceResponse.get("responseCode") != "SUCCESS":
        raise QualysAPIError(
            f"API response returned error: {serviceResponse.get('responseCode')}"
        )

    data = serviceResponse.get("data")
    if data.get("WebApp"):
        data = data.get("WebApp")
        return f"WebaApp {data} updated successfully."
    else:
        print("No data found. Exiting.")


def delete_webapp(
    auth: BasicAuth, removeFromSubscription: bool = True, **kwargs
) -> list[str]:
    """
    Delete webapps out of Qualys WAS.

    To delete a single asset, use kwarg 'id' set to
    a single int/str.

    To delete multiple assets,
    set 'id' to a list of int/str, or use the other
    kwargs/operators listed below.

    If both 'id' and other kwargs are provided, the 'id'
    kwarg will take precedence.

    Args:
        auth (BasicAuth): The authentication object.
        removeFromSubscription (bool): Whether to remove the webapp from the subscription (True) or just WAS (False). Default is True.

    ## Kwargs:

        - id (Union[str, int]): Web application ID.
        - id_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the ID filter.
        - name (str): Web application name.
        - name_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the name filter.
        - url (str): Web application URL.
        - url_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the URL filter.
        - tags_name (str): Tag name.
        - tags_name_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the tag name filter.
        - tags_id (Union[str, int]): Tag ID.
        - tags_id_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the tag ID filter.
        - createdDate (str): Date the web application was created in UTC date/time format.
        - createdDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the created date filter.
        - updatedDate (str): Date the web application was last updated in UTC date/time format.
        - updatedDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the updated date filter.
        - isScheduled (bool): Whether the web application has a scan scheduled.
        - isScheduled_operator (Literal["EQUALS", "NOT EQUALS"]): Operator for the isScheduled filter.
        - isScanned (bool): Whether the web application has been scanned.
        - isScanned_operator (Literal["EQUALS", "NOT EQUALS"]): Operator for the isScanned filter.
        - lastScan_status (Literal["SUBMITTED", "RUNNING", "FINISHED", "TIME_LIMIT_EXCEEDED", "SCAN_NOT_LAUNCHED", "SCANNER_NOT_AVAILABLE", "ERROR", "CANCELLED"]): Status of the last scan.
        - lastScan_status_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the last scan status filter.
        - lastScan_date (str): Date of the last scan in UTC date/time format.
        - lastScan_date_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the last scan date filter.

    Returns:
        list[str]: A list of IDs that were successfully deleted.
    """

    # Make sure user passed in a kwarg:
    if not kwargs:
        raise ValueError("No kwargs provided. Please provide at least one kwarg.")

    TO_STR = [
        "id",
        "tags_id",
    ]

    for key in TO_STR:
        if key in kwargs:
            kwargs[key] = str(kwargs[key])

    # Validate the kwargs:
    kwargs = validate_kwargs(endpoint="delete_webapp", **kwargs)

    # Build the XML payload:
    payload = build_service_request(**kwargs)
    payload["removeFromSubscription"] = removeFromSubscription

    # Make the API call:
    parsed = call_webapp_api(auth, "delete_webapp", payload)

    serviceResponse = parsed.get("ServiceResponse")
    if not serviceResponse:
        raise QualysAPIError("No ServiceResponse tag returned in the API response")

    if serviceResponse.get("responseCode") != "SUCCESS":
        raise QualysAPIError(
            f"API response returned error: {serviceResponse.get('responseCode')}"
        )

    if serviceResponse.get("count") == "0":
        print(f"No applicable web apps found. Exiting.")
        return []

    deleted = []
    data = serviceResponse.get("data")
    if data.get("WebApp"):
        data = data.get("WebApp")
        if isinstance(data, dict):
            data = [data]
        for webapp in data:
            deleted.append(webapp)

    return deleted


def purge_webapp(auth: BasicAuth, **kwargs) -> list[str]:
    """
    Purge webapp scan data out of Qualys WAS.

    To delete a single asset's scan history, use kwarg 'id' set to
    a single int/str.

    To delete multiple assets' scan history,
    set 'id' to a list of int/str, or use the other
    kwargs/operators listed below.

    If both 'id' and other kwargs are provided, the 'id'
    kwarg will take precedence.

    Args:
        auth (BasicAuth): The authentication object.

    ## Kwargs:

        - id (Union[str, int]): Web application ID.
        - id_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the ID filter.
        - name (str): Web application name.
        - name_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the name filter.
        - url (str): Web application URL.
        - url_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the URL filter.
        - tags_name (str): Tag name.
        - tags_name_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the tag name filter.
        - tags_id (Union[str, int]): Tag ID.
        - tags_id_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the tag ID filter.
        - createdDate (str): Date the web application was created in UTC date/time format.
        - createdDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the created date filter.
        - updatedDate (str): Date the web application was last updated in UTC date/time format.
        - updatedDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the updated date filter.
        - isScheduled (bool): Whether the web application has a scan scheduled.
        - isScheduled_operator (Literal["EQUALS", "NOT EQUALS"]): Operator for the isScheduled filter.
        - isScanned (bool): Whether the web application has been scanned.
        - isScanned_operator (Literal["EQUALS", "NOT EQUALS"]): Operator for the isScanned filter.
        - lastScan_status (Literal["SUBMITTED", "RUNNING", "FINISHED", "TIME_LIMIT_EXCEEDED", "SCAN_NOT_LAUNCHED", "SCANNER_NOT_AVAILABLE", "ERROR", "CANCELLED"]): Status of the last scan.
        - lastScan_status_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the last scan status filter.
        - lastScan_date (str): Date of the last scan in UTC date/time format.
        - lastScan_date_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the last scan date filter.

    Returns:
        list[str]: A list of IDs that had their scan data purged.
    """

    # Make sure user passed in a kwarg:
    if not kwargs:
        raise ValueError("No kwargs provided. Please provide at least one kwarg.")

    TO_STR = [
        "id",
        "tags_id",
    ]

    for key in TO_STR:
        if key in kwargs:
            kwargs[key] = str(kwargs[key])

    # Validate the kwargs. purge uses the same kwargs as delete:
    kwargs = validate_kwargs(endpoint="delete_webapp", **kwargs)

    # Build the XML payload:
    payload = build_service_request(**kwargs)

    # Make the API call:
    parsed = call_webapp_api(auth, "delete_webapp", payload)

    serviceResponse = parsed.get("ServiceResponse")
    if not serviceResponse:
        raise QualysAPIError("No ServiceResponse tag returned in the API response")

    if serviceResponse.get("responseCode") != "SUCCESS":
        raise QualysAPIError(
            f"API response returned error: {serviceResponse.get('responseCode')}"
        )

    if serviceResponse.get("count") == "0":
        print(f"No applicable web apps found. Exiting.")
        return []

    deleted = []
    data = serviceResponse.get("data")
    if data.get("WebApp"):
        data = data.get("WebApp")
        if isinstance(data, dict):
            data = [data]
        for webapp in data:
            deleted.append(webapp)

    return deleted


def get_selenium_script(
    auth: BasicAuth, id: Union[int, str], crawlingScripts_id: Union[int, str]
) -> object:
    """
    Download the associated Selenium script for a web application, identified by 1+ web app IDs.

    Args:
        auth (BasicAuth): The authentication object.
        id (Union[int, str]): The ID of the web application.
        crawlingScripts_id (Union[int, str]): The ID of the crawling script.

    Returns:
        object: The Selenium script.
    """

    # Validate the kwargs:
    validate_kwargs(
        endpoint="get_selenium_script",
        id=id,
        crawlingScripts_id=str(crawlingScripts_id),
    )

    # Format crawlingScripts ID:
    kwargs = {"id": str(id), "crawlingScripts.id": str(crawlingScripts_id)}

    # Build the XML payload:
    payload = build_service_request(**kwargs)

    # Make the API call:
    parsed = call_webapp_api(auth, "get_selenium_script", payload)

    serviceResponse = parsed.get("ServiceResponse")
    if not serviceResponse:
        raise QualysAPIError("No ServiceResponse tag returned in the API response")

    if serviceResponse.get("responseCode") != "SUCCESS":
        raise QualysAPIError(
            f"API response returned error: {serviceResponse.get('responseCode')}"
        )

    data = serviceResponse.get("data")
    print(
        "[WARNING] Code to parse data from this endpoint is not yet implemented. Please submit a PR. Returning raw data..."
    )
    return data
    # if data.get("SeleniumScript"):
    #    data = data.get("SeleniumScript")
    #    return data
    # else:
    #    print("No data found. Exiting.")



"""
Web Application Scanning (WAS) module
"""

from .webapps import (
    count_webapps,
    get_webapps,
    get_webapp_details,
    get_webapps_verbose,
    create_webapp,
    update_webapp,
    delete_webapp,
    purge_webapp,
    get_selenium_script,
)

from .authentication_records import (
    count_authentication_records,
    get_authentication_records,
)



"""
Code to generate XML service requests to Qualys WAS API
"""

from typing import Literal, Union, Dict, Any, List
from html import unescape
from re import compile

import xmltodict
from requests import Response

from ...base.xml_parser import xml_parser
from ...exceptions.Exceptions import QualysAPIError


def is_valid_regex(s: str) -> bool:
    """
    Attempts to use re.compile to test
    if a string is a valid regex.

    Args:
        s (str): The string to test.

    Returns:
        bool: Whether the string is a valid regex.
    """
    try:
        # Attempt to compile the string as a regex.
        compile(s)

        # If it contains common URL elements or non-regex patterns, return False
        # Examples: '://', '=', '&', 'www.', '.com', etc.
        if any(
            substring in s for substring in ["://", "=", "&", "www.", ".com", ".org"]
        ):
            return False

        return True
    except Exception as e:
        return False


def ensure_data_structure(request_dict: Dict[str, Any]) -> None:
    """
    Ensures that the 'data' and 'WebApp' keys exist in the request_dict.

    Args:
        request_dict (Dict[str, Any]): The dictionary where the 'data' and 'WebApp' keys need to be present.

    Returns:
        None
    """
    if "data" not in request_dict["ServiceRequest"]:
        request_dict["ServiceRequest"]["data"] = {}
    if "WebApp" not in request_dict["ServiceRequest"]["data"]:
        request_dict["ServiceRequest"]["data"]["WebApp"] = {}


def validate_list(
    input_data: Union[str, List[Any]], data_type: type, name: str
) -> List[Any]:
    """
    Validates that the input_data is a list of the specified data_type.
    Converts a comma-separated string into a list if needed.

    Args:
        input_data (Union[str, List[Any]]): The data to be validated, either as a list or a comma-separated string.
        data_type (type): The expected type of elements in the list (e.g., str, int).
        name (str): The name of the input (used for error messaging).

    Returns:
        List[Any]: The validated list of elements.

    Raises:
        ValueError: If input_data is not a list or a string that can be converted to a list of the expected type.
    """
    if isinstance(input_data, str):
        input_data = input_data.split(",")
    if not isinstance(input_data, list) or not all(
        isinstance(item, data_type) for item in input_data
    ):
        raise ValueError(
            f"{name} must be passed as a list of {data_type.__name__}s or a comma-separated string, not {input_data}"
        )
    return input_data


def update_request_dict(
    request_dict: Dict[str, Any],
    key: str,
    value: Any,
    action: Literal["add", "remove", "set"] = "set",
) -> None:
    """
    Updates the request_dict with a given key and value under the 'WebApp' section.

    Args:
        request_dict (Dict[str, Any]): The dictionary to be updated.
        key (str): The key under which the value will be stored.
        value (Any): The value to be stored in the request_dict.
        action (Literal["add", "remove", "set"], optional): The action to be taken on the value. Defaults to "set".

    Returns:
        None
    """
    ensure_data_structure(request_dict)
    request_dict["ServiceRequest"]["data"]["WebApp"][key] = {action: value}


def format_xml_list(items: List[Any], tag: str) -> str:
    """
    Formats a list of items into an XML string with the specified tag.

    Args:
        items (List[Any]): The list of items to be formatted as XML.
        tag (str): The XML tag to wrap around each item.

    Returns:
        str: A string containing the items formatted as XML elements.
    """
    if tag == "Tag":
        # Special case for tags, which have a different format
        return "".join(f"\n<Tag><id>{item}</id></Tag>" for item in items)
    else:
        return "".join(f"\n<{tag}>{item}</{tag}>" for item in items)


def validate_response(response: Response) -> dict:
    """
    Parse the XML response from the Qualys API
    and check for errors.

    Args:
        response (Response): The response object from the API call.

    Returns:
        dict: The parsed XML response.
    """
    parsed = xml_parser(response.text)

    if response.status_code != 200:
        errorMessage = parsed.get("ServiceResponse").get("responseErrorDetails")

        if errorMessage:
            errorMessage = errorMessage.get("errorMessage")

        responseCode = parsed.get("ServiceResponse").get("responseCode")
        raise QualysAPIError(
            f"Error retrieving web applications. Status code: {response.status_code}. Endpoint reporting: {responseCode}. Error message: {errorMessage}"
        )

    serviceResponse = parsed.get("ServiceResponse")
    if not serviceResponse:
        raise QualysAPIError("No ServiceResponse tag returned in the API response")

    if serviceResponse.get("responseCode") != "SUCCESS":
        errorMessage = (
            parsed.get("ServiceResponse")
            .get("responseErrorDetails")
            .get("errorMessage")
        )
        responseCode = parsed.get("ServiceResponse").get("responseCode")
        raise QualysAPIError(
            f"Error retrieving web applications. Status code: {response.status_code}. Endpoint reporting: {responseCode}. Error message: {errorMessage}"
        )

    return parsed


def build_service_request(
    _webapp_creation_or_edit: bool = False,
    action: Literal["set", "add", "remove"] = "set",
    authRecord_id: int = None,
    _uris: Union[str, list[str]] = None,
    tag_ids: Union[int, list[int]] = None,
    _domains: Union[str, list[str]] = None,
    _scannerTag_ids: Union[int, list[int]] = None,
    **kwargs,
) -> dict[str, str]:
    """
    Build the XML request. Users pass in the filters as kwargs.
    there is the filter itself as well as a filter_operator kwarg.
    If the filter_operator is not provided, default to "EQUALS".
    Do not create a Criteria tag for any _operator kwargs.

    Args:
        _webapp_creation_or_edit (bool, optional): If True, allows some tags to go into the DATA section instead of filters.
        action (Literal['set','add','remove'], optional): Whether to set, add, or remove various details.
        authRecord_id (int, optional): The ID of an authentication record
        _uris (Union[str, list[str]], optional): The URIs to be added to the WebApp. Can be a comma-separated string or a list of strings.
        tag_ids (Union[int, list[int]], optional): The tag IDs to be added to the WebApp. Can be a single integer or a list of integers.
        domains (Union[str, list[str]], optional): The domains to be added to the WebApp. Can be a single string or a list of strings.
        _scannerTag_ids (Union[int, list[int]], optional): A tag ID associated with 1+ scanners to assign to the WebApp.

    Returns:
        dict: The XML payload to be sent to the Qualys API.
    """

    request_dict = {"ServiceRequest": {"filters": {}}}

    # Check for any tags that need to be under preferences instead of filters:
    PREFERENCE_TAGS = ["verbose"]

    # If this is a webapp creation or edit, move some tags to the DATA section:
    DATA_TAGS = ["name", "url"]

    preferences = {}
    for tag in PREFERENCE_TAGS:
        if tag in kwargs:
            preferences[tag] = (
                str(kwargs.pop(tag)).lower()
                if isinstance(kwargs[tag], bool)
                else kwargs.pop(tag)
            )

    if preferences:
        request_dict["ServiceRequest"]["preferences"] = preferences

    if _webapp_creation_or_edit:
        data = {"WebApp": {}}
        for tag in DATA_TAGS:
            if tag in kwargs:
                data["WebApp"][tag] = (
                    str(kwargs.pop(tag)).lower()
                    if isinstance(kwargs[tag], bool)
                    else kwargs.pop(tag)
                )
        request_dict["ServiceRequest"]["data"] = data

    # Integrate authRecord_id into the main XML structure if provided
    if authRecord_id:
        update_request_dict(
            request_dict, "authRecords", {"WebAppAuthRecord": {"id": authRecord_id}}
        )

    # Integrate URIs into the main XML structure if provided
    if _uris:
        uris = validate_list(_uris, str, "URIs")
        uriList = format_xml_list(uris, "Url")
        update_request_dict(request_dict, "uris", uriList)

    # Integrate tag_ids into the main XML structure if provided
    if tag_ids:
        tag_ids = validate_list(tag_ids, int, "Tag IDs")
        tagList = format_xml_list(tag_ids, "Tag")
        update_request_dict(request_dict, "tags", tagList)

    # Integrate domains into the main XML structure if provided
    if _domains:
        domains = validate_list(_domains, str, "Domains")
        domainList = format_xml_list(domains, "Domain")
        update_request_dict(request_dict, "domains", domainList)

    if _scannerTag_ids:
        scannerTag_ids = validate_list(_scannerTag_ids, int, "Scanner Tag IDs")
        scannerTagList = format_xml_list(scannerTag_ids, "Tag")
        update_request_dict(request_dict, "defaultScannerTags", scannerTagList)

    # Build the Criteria tags for the filters:
    filters = []
    for kwarg, value in kwargs.items():
        if kwarg.endswith(".operator"):
            continue
        # fields that go under WebApp:
        if kwarg in [
            "authRecord.id",
            "uris",
            "tag.ids",
            "domains",
            "scannerTag.ids",
            "removeFromSubscription",
        ]:
            continue
        criteria = {
            "@field": kwarg,
            "@operator": kwargs.get(f"{kwarg}.operator", "EQUALS"),
            "#text": str(value).lower() if isinstance(value, bool) else value,
        }
        filters.append(criteria)

    if filters:
        request_dict["ServiceRequest"]["filters"]["Criteria"] = filters
        # If any filters snuck in that have a None #text value, remove them
        for criteria in request_dict["ServiceRequest"]["filters"]["Criteria"]:
            if criteria["#text"] is None:
                request_dict["ServiceRequest"]["filters"]["Criteria"].remove(criteria)
    else:
        del request_dict["ServiceRequest"]["filters"]

    # Unescape any HTML. Necessary due to xmltodict's behavior.
    try:
        xml = unescape(xmltodict.unparse(request_dict, pretty=True))
        payload = {"_xml_data": xml}
    except Exception as e:
        raise ValueError(f"Error unescaping XML: {e}")

    return payload


def build_update_request(
    name: str = None,
    url: str = None,
    attributes: Dict[
        Literal["add"] : [{"key", "value"}], Literal["remove"] : [str]
    ] = None,
    defaultProfile_id: int = None,
    _urlExcludelist: List[str] = None,
    _urlAllowlist: List[str] = None,
    _postDataExcludelist: List[str] = None,
    _useSitemap: bool = None,
    _headers: List[str] = None,
    _authRecord_id: Dict[Literal["add"] : [], Literal["remove"] : []] = None,
    **kwargs,
) -> dict[str, str]:
    """
    Build the XML update request. Users pass in the filters as kwargs.
    there is the filter itself as well as a filter_operator kwarg.
    If the filter_operator is not provided, default to "EQUALS".

    Args:
        name (str, optional): The name of the web application.
        url (str, optional): The URL of the web application.
        attributes (Dict[Literal['add']: List[Dict[str, str]], Literal['remove']: List[str]], optional): The attributes to add or remove from the web application.
        defaultProfile_id (int, optional): The ID of the default profile to assign to the web application.
        _urlExcludelist (List[str], optional): The list of URLs to exclude from the web application.
        _urlAllowlist (List[str], optional): The list of URLs to allow in the web application.
        _postDataExcludelist (List[str], optional): The list of post data to exclude from the web application.
        _useSitemap (bool, optional): Whether to use the sitemap for the web application.
        _headers (List[str], optional): The headers to add to the web application.
        _authRecord_id (Dict[Literal['add']: List[int], Literal['remove']: List[int]], optional): The authentication record IDs to add or remove from the web application.
        kwargs (dict): Additional filters to be passed to the Qualys API.

    Returns:
        dict: The XML payload to be sent to the Qualys API.
    """

    request_dict = {"ServiceRequest": {"data": {"WebApp": {}}}}

    if name:
        request_dict["ServiceRequest"]["data"]["WebApp"]["name"] = name

    if url:
        request_dict["ServiceRequest"]["data"]["WebApp"]["url"] = url

    if defaultProfile_id:
        request_dict["ServiceRequest"]["data"]["WebApp"]["defaultProfile"] = {
            "id": defaultProfile_id
        }

    if attributes:
        add_data = attributes.get("add")
        remove_data = attributes.get("remove")
        if add_data and not isinstance(add_data, list):
            add_data = [add_data]
        # Ensure that everything in the list is a dictionary
        for item in add_data:
            if not isinstance(item, dict):
                raise ValueError(
                    f"Attributes must be passed as a list of dictionaries. Key-value pairs are expected to add attributes. Received: {add_data}"
                )
        if remove_data and not isinstance(remove_data, list):
            remove_data = [remove_data]
        if add_data:
            for itm in add_data:
                for k, v in itm.items():
                    request_dict["ServiceRequest"]["data"]["WebApp"]["attributes"] = {
                        "update": {"Attribute": {"name": k, "value": v}}
                    }
        if remove_data:
            request_dict["ServiceRequest"]["data"]["WebApp"]["attributes"] = {
                "remove": {"Attribute": {"name": remove_data}}
            }

    if _authRecord_id:
        add_data = _authRecord_id.get("add")
        remove_data = _authRecord_id.get("remove")
        if add_data and not isinstance(add_data, list):
            add_data = [add_data]
        if remove_data and not isinstance(remove_data, list):
            remove_data = [remove_data]
        if add_data:
            request_dict["ServiceRequest"]["data"]["WebApp"]["authRecords"] = {
                "add": {"WebAppAuthRecord": {"id": add_data}}
            }
        if remove_data:
            request_dict["ServiceRequest"]["data"]["WebApp"]["authRecords"] = {
                "remove": {"WebAppAuthRecord": {"id": remove_data}}
            }

    if _urlExcludelist:
        urlExcludelist = _urlExcludelist
        if not isinstance(urlExcludelist, list):
            raise ValueError(
                f"urlExcludelist must be passed as a list of strings. Received: {urlExcludelist}"
            )

        l = []
        for entry in urlExcludelist:
            # Check for any regex characters in the entry.
            # If found, set the regex attribute to true.
            if is_valid_regex(entry):
                l.append(
                    xmltodict.unparse(
                        {"UrlEntry": {"@regex": "true", "#text": entry}},
                        full_document=False,
                    )
                )
            else:
                l.append(
                    xmltodict.unparse(
                        {"UrlEntry": {"@regex": "false", "#text": entry}},
                        full_document=False,
                    )
                )

        request_dict["ServiceRequest"]["data"]["WebApp"]["urlExcludelist"] = {
            "set": "\n".join(l)
        }

    if _urlAllowlist:
        if not isinstance(_urlAllowlist, list):
            raise ValueError(
                f"urlAllowlist must be passed as a list of strings. Received: {_urlAllowlist}"
            )
        l = []
        for entry in _urlAllowlist:
            # Check for any regex characters in the entry.
            # If found, set the regex attribute to true.
            if is_valid_regex(entry):
                l.append(
                    xmltodict.unparse(
                        {"UrlEntry": {"@regex": "true", "#text": entry}},
                        full_document=False,
                    )
                )
            else:
                l.append(
                    xmltodict.unparse(
                        {"UrlEntry": {"@regex": "false", "#text": entry}},
                        full_document=False,
                    )
                )

        request_dict["ServiceRequest"]["data"]["WebApp"]["urlAllowlist"] = {
            "set": "\n".join(l)
        }

    if _postDataExcludelist:
        if not isinstance(_postDataExcludelist, list):
            raise ValueError(
                f"postDataExcludelist must be passed as a list of strings. Received: {_postDataExcludelist}"
            )

        l = []
        for entry in _postDataExcludelist:
            # Check for any regex characters in the entry.
            # If found, set the regex attribute to true.
            if is_valid_regex(entry):
                l.append(
                    xmltodict.unparse(
                        {"UrlEntry": {"@regex": "true", "#text": entry}},
                        full_document=False,
                    )
                )
            else:
                l.append(
                    xmltodict.unparse(
                        {"UrlEntry": {"@regex": "false", "#text": entry}},
                        full_document=False,
                    )
                )

        request_dict["ServiceRequest"]["data"]["WebApp"]["postDataExcludelist"] = {
            "set": "\n".join(l)
        }

    if _useSitemap is not None:
        request_dict["ServiceRequest"]["data"]["WebApp"]["useSitemap"] = str(
            _useSitemap
        ).lower()

    if _headers:
        if not isinstance(_headers, list):
            raise ValueError(
                f"headers must be passed as a list of strings. Received: {_headers}"
            )
        l = []
        for header in _headers:
            l.append(xmltodict.unparse({"WebAppHeader": header}, full_document=False))
        request_dict["ServiceRequest"]["data"]["WebApp"]["headers"] = {
            "set": "\n".join(l)
        }

    # Unescape any HTML. Necessary due to xmltodict's behavior.
    try:
        xml = unescape(xmltodict.unparse(request_dict, pretty=True))
        payload = {"_xml_data": xml}
    except Exception as e:
        raise ValueError(f"Error unescaping XML: {e}")

    return payload



"""
query_kb.py - contains the query_kb function for the qualysdk package.

This function is used to query the Qualys KnowledgeBase (KB), which is a database of vulnerabilities and their details.
"""

from typing import *
from urllib.parse import parse_qs, urlparse

from .data_classes.kb_entry import KBEntry
from .data_classes.qvs import KBQVS
from ..base.base_list import BaseList
from ..base.call_api import call_api
from ..auth.token import BasicAuth
from ..base.xml_parser import xml_parser


def query_kb(auth: BasicAuth, **kwargs) -> BaseList[KBEntry]:
    """
    Query the Qualys KnowledgeBase (KB) for vulnerabilities matching the kiven kwargs.
    Can be used to download the entire KB (no kwargs) or to search for specific vulnerabilities.
    NOTE: this function automatically appends 'action=list' to the kwargs. You do not need to include it. Should you do so, it will be overwritten. It is just recognized as valid for the sake of completeness.

    Params:
        auth (BasicAuth) The authentication object.

    :Kwargs:
        action (str) The action to perform. Default is 'list'. WARNING: any value you pass is overwritten with 'list'. It is just recognized as valid for the sake of completeness.
        code_modified_after (str): The date to search for vulnerabilities modified after Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        code_modified_before (str): The date to search for vulnerabilities modified before Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        echo_request (str): Response will include the request you sent.
        details Literal["Basic", "All", "None"]: The level of detail to return. Default is 'Basic'.
        ids (str): The IDs of the vulnerabilities to return as a comma-separated string.
        id_min (int): The minimum ID of the vulnerabilities to return.
        id_max (int): The maximum ID of the vulnerabilities to return.
        is_patchable (bool): Whether the vulnerability is patchable. Default is 'False'.
        last_modified_after (str): The date to search for vulnerabilities last modified after Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        last_modified_before (str): The date to search for vulnerabilities last modified before Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        last_modified_by_user_after (str): The date to search for vulnerabilities last modified by user after Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        last_modified_by_user_before (str): The date to search for vulnerabilities last modified by user before Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        last_modified_by_service_after (str): The date to search for vulnerabilities last modified by service after Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        last_modified_by_service_before (str): The date to search for vulnerabilities last modified by service before Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        published_after (str): The date to search for vulnerabilities published after Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        published_before (str): The date to search for vulnerabilities published before Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        discovery_method (str): The discovery method of the vulnerability.
        discovery_auth_types (str): The authentication types used to discover the vulnerability.
        show_pci_reasons (bool): Whether to show PCI reasons. Default is 'False'.
        show_supported_modules_info (bool): Whether to show supported modules info. Default is 'False'.
        show_disabled_flag (bool): Whether to show the disabled flag. Default is 'False'.
        show_qid_change_log (bool): Whether to show the QID change log. Default is 'False'.

    Returns:
        BaseList of KBEntry objects representing the vulnerabilities returned by the query.
    """

    # add the action to the kwargs:
    kwargs["action"] = "list"

    responses = BaseList()
    pulled = 0

    # qualys expects all boolean values to be represented as a 0 or 1:
    for key, value in kwargs.items():
        if isinstance(value, bool):
            kwargs[key] = 1 if value else 0

    while True:
        # make the request:
        response = call_api(
            auth=auth,
            module="vmdr",
            endpoint="query_kb",
            params=kwargs,
            headers={"X-Requested-With": "qualysdk SDK"},
        )
        if response.status_code != 200:
            raise Exception(f"Error: {response.status_code} - {response.text}")

        xml = xml_parser(response.text)

        # check if there is no vuln list
        if "VULN_LIST" not in xml["KNOWLEDGE_BASE_VULN_LIST_OUTPUT"]["RESPONSE"]:
            break

        for e in xml["KNOWLEDGE_BASE_VULN_LIST_OUTPUT"]["RESPONSE"]["VULN_LIST"][
            "VULN"
        ]:
            responses.append(KBEntry.from_dict(e))  # append entry

        pulled += 1
        print(f"Page {pulled} complete.")
        # KB API normally does not paginate, but if it does
        if "WARNING" in xml["KNOWLEDGE_BASE_VULN_LIST_OUTPUT"]["RESPONSE"]:
            if "URL" in xml["KNOWLEDGE_BASE_VULN_LIST_OUTPUT"]["RESPONSE"]["WARNING"]:
                print(
                    f"Pagination detected. Pulling next page from url: {xml['KNOWLEDGE_BASE_VULN_LIST_OUTPUT']['RESPONSE']['WARNING']['URL']}"
                )
                # parse the url to get the query params
                ps = parse_qs(
                    urlparse(
                        xml["KNOWLEDGE_BASE_VULN_LIST_OUTPUT"]["RESPONSE"]["WARNING"][
                            "URL"
                        ]
                    ).query
                )
                # update the kwargs with the new params
                kwargs.update(ps)

            else:
                break
        else:
            break

    return responses


def get_kb_qvs(auth: BasicAuth, cve: str, **kwargs) -> BaseList[KBQVS]:
    """
    Download Qualys KB QVS (Qualys Vulnerability Score) data for 1+ CVEs.

    Params:
        auth (BasicAuth) The authentication object.
        cve (str): The CVE ID(s) to search for. Comma-separated string.
        **kwargs: Additional filters/parameters to pass to the API. See below for details.

    :Kwargs:
        action (Literal['list']): The action to perform. Default is 'list'. WARNING: any value you pass is overwritten with 'list'. It is just recognized as valid for the sake of completeness.
        details (Literal['All', 'Basic', 'None']): The level of detail to return. Default is 'Basic'.
        qvs_last_modified_before (str): The date to search for QVS last modified before Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        qvs_last_modified_after (str): The date to search for QVS last modified after Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        qvs_min (int): The minimum QVS score to return.
        qvs_max (int): The maximum QVS score to return.
        nvd_published_before (str): The date to search for NVD published before Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        nvd_published_after (str): The date to search for NVD published after Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.

    Returns:
        BaseList of KBQVS objects representing the QVS data for the CVEs returned by the query.
    """

    kwargs["action"] = "list"
    if kwargs.get("details"):
        kwargs["details"] = kwargs["details"].capitalize()
    kwargs["cve"] = cve

    bl = BaseList()

    response = call_api(
        auth=auth,
        module="vmdr",
        endpoint="get_kb_qvs",
        params=kwargs,
        headers={"X-Requested-With": "qualysdk SDK"},
    )

    if r := response.json():
        for data in r.values():
            bl.append(KBQVS.from_dict(data))

    return bl
